/*
 * Rjb - Ruby <-> Java Bridge
 * Copyright(c) 2004 arton
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * $Id:$
 * $Log:$
 */

#define RJB_VERSION "0.0.5"

#include "ruby.h"
#include "st.h"
#include <jni.h>

#if !defined(COUNTOF)
#define COUNTOF(x) (sizeof(x)/sizeof(x[0]))
#endif

typedef void (*R2J)(VALUE, jvalue*, const char*, int);
typedef VALUE (*J2R)(jvalue);
typedef jarray (*R2JARRAY)(VALUE, const char*);
typedef jvalue (JNICALL *INVOKEA)(JNIEnv*, jobject, jmethodID, const jvalue*);

/*
 * Method Modifier Flag defined in
 * http://java.sun.com/docs/books/vmspec/2nd-edition/html/ClassFile.doc.html#88358
 */
#define ACC_PUBLIC  0x0001
#define ACC_PRIVATE  0x0002
#define ACC_PROTECTED  0x0004
#define ACC_STATIC  0x0008
#define ACC_FINAL  0x0010
#define ACC_VOLATILE  0x0040
#define ACC_TRANSIENT  0x0080

/*
 * internal method class
 */
struct cls_constructor {
    jmethodID id;
    int arg_count;
    R2J* arg_convert;
    char* method_signature;	
};

struct cls_method {
    struct cls_constructor basic;
    ID name;
    int static_method;
    off_t method;
    J2R result_convert;
    /* overload only */
    struct cls_method* next;
};

/*
 * Object instance
 */
struct jvi_data {
    jclass klass; /* class */
    jobject obj; /* instance */
    st_table* methods;
};

/*
 * Class instance
 */
struct jv_data {
    struct jvi_data idata;
    st_table* static_methods;
    struct cls_constructor** constructors;
};

static void register_class(VALUE, VALUE);
static VALUE register_instance(struct jvi_data*, jobject);
static VALUE rjb_s_free(struct jv_data*);
static void setup_metadata(struct jv_data*);

static VALUE rjb;
static VALUE rjbc;
static VALUE rjbi;
static VALUE loaded_classes;
static JavaVM* jvm;
static JNIEnv* jenv;

/*
 * Object cache, never destroyed
 */
static jclass j_method;
static jmethodID method_getModifiers;
static jmethodID method_getName;
static jmethodID getParameterTypes;
static jmethodID getReturnType;

static jclass j_constructor;
static jmethodID ctrGetParameterTypes;

static jclass j_class;
static jmethodID class_getName;

static jclass j_string;
static jmethodID str_tostring;

static jclass j_object;

/*
 * Type convertion tables
 */
static const char* jtype_tbl[] = {
  "byte", "char", "double", "float", "int", "long", "short", "boolean", "void",
  "java.lang.String",
};

static const char* jtype_etbl[] = {
    "B", "C", "D", "F", "I", "J", "S", "Z", "V",
    "Ljava.lang.String;",
};

static char* next_sig(char* p)
{
    if (!*p)
    {
	return p;
    }
    while (!strchr("BCDEFIJSZV;", *p))
    {
	p++;
    }
    return ++p;
}

/*
 * handle Java exception
 */
static void check_exception(int t)
{
    jthrowable exp = (*jenv)->ExceptionOccurred(jenv);
    if (exp)
    {
	(*jenv)->ExceptionDescribe(jenv);
	(*jenv)->ExceptionClear(jenv);
        if (t)
	{
	    rb_raise(rb_eRuntimeError, "exception");
        }
    }
}

static VALUE jstring2val(jstring s)
{
    const char* p = (*jenv)->GetStringUTFChars(jenv, s, JNI_FALSE);
    VALUE v = rb_str_new2(p);
    (*jenv)->ReleaseStringUTFChars(jenv, s, p);
    (*jenv)->DeleteLocalRef(jenv, s);
    return v;
}

/*
 * conversion methods
 */
static VALUE jv2rv(jvalue val)
{
    const char* cname;
    jstring nm;
    jclass klass;
    VALUE clsname;
    VALUE v;
    struct jv_data* ptr;
    // object to ruby
    if (!val.l) return Qnil;
    klass = (*jenv)->GetObjectClass(jenv, val.l);
    nm = (*jenv)->CallObjectMethod(jenv, klass, class_getName);
    check_exception(0);
    cname = (*jenv)->GetStringUTFChars(jenv, nm, JNI_FALSE);
    clsname = rb_str_new2(cname);
    (*jenv)->ReleaseStringUTFChars(jenv, nm, cname);
    (*jenv)->DeleteLocalRef(jenv, nm);
    v = rb_hash_aref(loaded_classes, clsname);
    if (v == Qnil)
    {
	ptr = ALLOC(struct jv_data);
	memset(ptr, 0, sizeof(struct jv_data));
	v = Data_Wrap_Struct(rjbc, NULL, rjb_s_free, ptr);
	ptr->idata.obj = klass;
	setup_metadata(ptr);
	register_class(v, clsname);
    }
    Data_Get_Struct(v, struct jv_data, ptr);
    return register_instance((struct jvi_data*)ptr, val.l);
}

static VALUE jvoid2rv(jvalue val)
{
    return Qnil;
}

static VALUE jbyte2rv(jvalue val)
{
    return INT2NUM(val.b);
}

static VALUE jchar2rv(jvalue val)
{
    return INT2NUM(val.c);
}

static VALUE jdouble2rv(jvalue val)
{
    return rb_float_new(val.d);
}

static VALUE jfloat2rv(jvalue val)
{
    return rb_float_new(val.f);
}

static VALUE jint2rv(jvalue val)
{
    return INT2NUM(val.i);    
}

static VALUE jlong2rv(jvalue val)
{
#if HAVE_LONG_LONG
    return LL2NUM(val.j);
#else
    return Qnil;
#endif
}

static VALUE jshort2rv(jvalue val)
{
    return INT2NUM(val.s);
}

static VALUE jboolean2rv(jvalue val)
{
    return (val.z) ? Qtrue : Qfalse;
}

static VALUE jstring2rv(jvalue val)
{
    return jstring2val((jstring)val.l);
}

static VALUE jarray2rv(jvalue val)
{
    jsize len;
    VALUE v;
    int i;
    if (!val.l) return Qnil;
    len = (*jenv)->GetArrayLength(jenv, val.l);
    v = rb_ary_new2(len);
    for (i = 0; i < len; i++)
    {
	jvalue wrap;
	wrap.l = (*jenv)->GetObjectArrayElement(jenv, val.l, i);
	rb_ary_push(v, jv2rv(wrap));
    }
    (*jenv)->DeleteLocalRef(jenv, val.l);
    return v;
}

static VALUE jbytearray2rv(jvalue val)
{
    jsize len = (*jenv)->GetArrayLength(jenv, val.l);
    jbyte* p = (*jenv)->GetByteArrayElements(jenv, val.l, NULL);
    VALUE v = rb_str_new((char*)p, len);
    (*jenv)->ReleaseByteArrayElements(jenv, val.l, p, JNI_ABORT);
    (*jenv)->DeleteLocalRef(jenv, val.l);
    return v;
}

static VALUE jchararray2rv(jvalue val)
{
    int i;
    jsize len = (*jenv)->GetArrayLength(jenv, val.l);
    jchar* p = (*jenv)->GetCharArrayElements(jenv, val.l, NULL);
    VALUE v = rb_ary_new2(len);
    VALUE* pv = RARRAY(v)->ptr;
    RARRAY(v)->len = len;
    for (i = 0; i < len; i++)
    {
	*pv++ = INT2FIX(*(p + i));
    }
    (*jenv)->ReleaseCharArrayElements(jenv, val.l, p, JNI_ABORT);
    (*jenv)->DeleteLocalRef(jenv, val.l);
    return v;
}
static VALUE jdoublearray2rv(jvalue val)
{
    int i;
    jsize len = (*jenv)->GetArrayLength(jenv, val.l);
    jdouble* p = (*jenv)->GetDoubleArrayElements(jenv, val.l, NULL);
    VALUE v = rb_ary_new2(len);
    VALUE* pv = RARRAY(v)->ptr;
    RARRAY(v)->len = len;
    for (i = 0; i < len; i++)
    {
	*pv++ = rb_float_new(*(p + i));
    }
    (*jenv)->ReleaseDoubleArrayElements(jenv, val.l, p, JNI_ABORT);
    (*jenv)->DeleteLocalRef(jenv, val.l);
    return v;
}
static VALUE jfloatarray2rv(jvalue val)
{
    int i;
    jsize len = (*jenv)->GetArrayLength(jenv, val.l);
    jfloat* p = (*jenv)->GetFloatArrayElements(jenv, val.l, NULL);
    VALUE v = rb_ary_new2(len);
    VALUE* pv = RARRAY(v)->ptr;
    RARRAY(v)->len = len;
    for (i = 0; i < len; i++)
    {
	*pv++ = rb_float_new(*(p + i));
    }
    (*jenv)->ReleaseFloatArrayElements(jenv, val.l, p, JNI_ABORT);
    (*jenv)->DeleteLocalRef(jenv, val.l);
    return v;
}
static VALUE jintarray2rv(jvalue val)
{
    int i;
    jsize len = (*jenv)->GetArrayLength(jenv, val.l);
    jint* p = (*jenv)->GetIntArrayElements(jenv, val.l, NULL);
    VALUE v = rb_ary_new2(len);
    VALUE* pv = RARRAY(v)->ptr;
    RARRAY(v)->len = len;
    for (i = 0; i < len; i++)
    {
	*pv++ = INT2NUM(*(p + i));
    }
    (*jenv)->ReleaseIntArrayElements(jenv, val.l, p, JNI_ABORT);
    (*jenv)->DeleteLocalRef(jenv, val.l);
    return v;
}
static VALUE jlongarray2rv(jvalue val)
{
    int i;
    jsize len = (*jenv)->GetArrayLength(jenv, val.l);
    jlong* p = (*jenv)->GetLongArrayElements(jenv, val.l, NULL);
    VALUE v = rb_ary_new2(len);
    VALUE* pv = RARRAY(v)->ptr;
    RARRAY(v)->len = len;
    for (i = 0; i < len; i++)
    {
#if HAVE_LONG_LONG
	*pv++ = LL2NUM(*(p + i));
#else
	*pv++ = LONG2NUM(*(p + i));
#endif	
    }
    (*jenv)->ReleaseLongArrayElements(jenv, val.l, p, JNI_ABORT);
    (*jenv)->DeleteLocalRef(jenv, val.l);
    return v;
}
static VALUE jshortarray2rv(jvalue val)
{
    int i;
    jsize len = (*jenv)->GetArrayLength(jenv, val.l);
    jshort* p = (*jenv)->GetShortArrayElements(jenv, val.l, NULL);
    VALUE v = rb_ary_new2(len);
    VALUE* pv = RARRAY(v)->ptr;
    RARRAY(v)->len = len;
    for (i = 0; i < len; i++)
    {
	*pv++ = INT2FIX(*(p + i));
    }
    (*jenv)->ReleaseShortArrayElements(jenv, val.l, p, JNI_ABORT);
    (*jenv)->DeleteLocalRef(jenv, val.l);
    return v;
}
static VALUE jstringarray2rv(jvalue val)
{
    int i;
    jsize len = (*jenv)->GetArrayLength(jenv, val.l);
    VALUE v = rb_ary_new2(len);
    for (i = 0; i < len; i++)
    {
	jobject p = (*jenv)->GetObjectArrayElement(jenv, val.l, i);
	rb_ary_push(v, jstring2val((jstring)p));
    }
    (*jenv)->DeleteLocalRef(jenv, val.l);
    return v;
}
static VALUE jbooleanarray2rv(jvalue val)
{
    int i;
    jsize len = (*jenv)->GetArrayLength(jenv, val.l);
    jboolean* p = (*jenv)->GetBooleanArrayElements(jenv, val.l, NULL);
    VALUE v = rb_ary_new2(len);
    VALUE* pv = RARRAY(v)->ptr;
    RARRAY(v)->len = len;
    for (i = 0; i < len; i++)
    {
	*pv++ = (*(p + i)) ? Qtrue : Qfalse;
    }
    (*jenv)->ReleaseBooleanArrayElements(jenv, val.l, p, JNI_ABORT);
    (*jenv)->DeleteLocalRef(jenv, val.l);
    return v;
}

static void rv2jv(VALUE val, jvalue* jv, const char* psig, int release)
{
    jv->l = NULL;
}

static void rv2jbyte(VALUE val, jvalue* jv, const char* psig, int release)
{
    if (!release)
	jv->b = (jbyte)NUM2INT(val);
}
static void rv2jchar(VALUE val, jvalue* jv, const char* psig, int release)
{
    if (!release)
	jv->c = (jchar)NUM2INT(val);
}
static void rv2jdouble(VALUE val, jvalue* jv, const char* psig, int release)
{
    if (release) return;
    switch (TYPE(val))
    {
    case T_FIXNUM:
	jv->d = NUM2INT(val);
	break;
    case T_FLOAT:
	jv->d = NUM2DBL(val);
	break;
    default:
	rb_raise(rb_eRuntimeError, "can't change to double");
	break;
    }
}
static void rv2jfloat(VALUE val, jvalue* jv, const char* psig, int release)
{
    if (release) return;
    switch (TYPE(val))
    {
    case T_FIXNUM:
	jv->f = (float)NUM2INT(val);
	break;
    case T_FLOAT:
	jv->f = (float)NUM2DBL(val);
	break;
    default:
	rb_raise(rb_eRuntimeError, "can't change to float");
	break;
    }
}
static void rv2jint(VALUE val, jvalue* jv, const char* psig, int release)
{
    if (!release)
	jv->i = NUM2INT(val);    
}
static void rv2jlong(VALUE val, jvalue* jv, const char* psig, int release)
{
    if (release) return;
    switch (TYPE(val))
    {
    case T_FIXNUM:
	jv->j = NUM2INT(val);
	break;
    default:
#if HAVE_LONG_LONG
	jv->j = NUM2LL(val);
#else
	rb_raise(rb_eRuntimeError, "can't change to long");
#endif
	break;
    }
}
static void rv2jshort(VALUE val, jvalue* jv, const char* psig, int release)
{
    if (!release)
	jv->s = (short)NUM2INT(val);    
}
static void rv2jboolean(VALUE val, jvalue* jv, const char* psig, int release)
{
    if (!release)
	jv->z = (val == Qnil || val == Qfalse) ? JNI_FALSE : JNI_TRUE;
}
static void rv2jstring(VALUE val, jvalue* jv, const char* psig, int release)
{
    if (!release)
    {
	if (TYPE(val) == T_DATA
	    && (RBASIC(val)->klass == rjbi || RBASIC(val)->klass == rjb))
	{
	    struct jvi_data* ptr;
	    Data_Get_Struct(val, struct jvi_data, ptr);
	    if ((*jenv)->IsInstanceOf(jenv, ptr->obj, j_string))
	    {
		jv->l = ptr->obj;
	    }
	    else
	    {
		jmethodID tostr;
		jstring js;
		tostr = (*jenv)->GetMethodID(jenv, ptr->klass, "toString", "()Ljava/lang/String;");
		check_exception(0);
		js = (*jenv)->CallObjectMethod(jenv, ptr->obj, tostr);
		check_exception(0);
		jv->l = js;
	    }
	}
	else
	{
	    char* p = StringValueCStr(val);
	    jstring js = (*jenv)->NewStringUTF(jenv, p);
	    jv->l = js;
	}
    }
    else
    {
	if (TYPE(val) == T_DATA
	    && (RBASIC(val)->klass == rjbi || RBASIC(val)->klass == rjb))
	{
	    struct jvi_data* ptr;
	    Data_Get_Struct(val, struct jvi_data, ptr);
	    if ((*jenv)->IsInstanceOf(jenv, ptr->obj, j_string))
	    {
		return; // never delete at this time
	    }
	}
	(*jenv)->DeleteLocalRef(jenv, jv->l);
    }
}
static void rv2jobject(VALUE val, jvalue* jv, const char* psig, int release)
{
    if (!release)
    {
	jv->l = NULL;
	if (TYPE(val) == T_DATA
	    && (RBASIC(val)->klass == rjbi || RBASIC(val)->klass == rjb))
	{
	    // TODO: check instanceof (class (in psig) )
	    struct jvi_data* ptr;
	    Data_Get_Struct(val, struct jvi_data, ptr);
	    jv->l = ptr->obj;
	}
	// TODO: ruby object -> Java Object
    }
}

static const R2J r2j[] = {
    rv2jbyte, rv2jchar, rv2jdouble, rv2jfloat, rv2jint, rv2jlong, rv2jshort,
    rv2jboolean, rv2jv, rv2jstring,
};

static const J2R j2r[] = {
    jbyte2rv, jchar2rv, jdouble2rv, jfloat2rv, jint2rv, jlong2rv, jshort2rv,
    jboolean2rv, jvoid2rv, jstring2rv,
};

static void check_fixnumarray(VALUE v)
{
    int i;
    int len = RARRAY(v)->len;
    VALUE* p = RARRAY(v)->ptr;
    // check all fixnum (overflow is permit)
    for (i = 0; i < len; i++)
    {
	if (!FIXNUM_P(*p++))
	{
	    rb_raise(rb_eRuntimeError, "array element must be a fixnum");
	}
    }
}

static jarray r2barray(VALUE v, const char* cls)
{
    jarray ary = NULL;
    if (TYPE(v) == T_STRING)
    {
	ary = (*jenv)->NewByteArray(jenv, RSTRING(v)->len);
	(*jenv)->SetByteArrayRegion(jenv, ary, 0, RSTRING(v)->len,
				    RSTRING(v)->ptr);
    }
    else if (TYPE(v) == T_ARRAY)
    {
	int i;
	jbyte* pb;
	check_fixnumarray(v);
	ary = (*jenv)->NewByteArray(jenv, RARRAY(v)->len);
	pb = (*jenv)->GetByteArrayElements(jenv, ary, NULL);
	for (i = 0; i < RARRAY(v)->len; i++)
	{
	    *(pb + i) = (jbyte)FIX2ULONG(RARRAY(v)->ptr[i]);
	}
	(*jenv)->ReleaseByteArrayElements(jenv, ary, pb, 0);
    }
    if (!ary)
    {
	rb_raise(rb_eRuntimeError, "can't coerce to byte array");
    }
    return ary;
}

static jarray r2carray(VALUE v, const char* cls)
{
    jarray ary = NULL;
    if (TYPE(v) == T_ARRAY)
    {
	int i;
	jchar* pb;
	check_fixnumarray(v);
	ary = (*jenv)->NewCharArray(jenv, RARRAY(v)->len);
	pb = (*jenv)->GetCharArrayElements(jenv, ary, NULL);
	for (i = 0; i < RARRAY(v)->len; i++)
	{
	    *(pb + i) = (jchar)FIX2ULONG(RARRAY(v)->ptr[i]);
	}
	(*jenv)->ReleaseCharArrayElements(jenv, ary, pb, 0);
	return ary;
    }
    rb_raise(rb_eRuntimeError, "can't coerce to char array");
}

static jarray r2darray(VALUE v, const char* cls)
{
    jarray ary = NULL;
    if (TYPE(v) == T_ARRAY)
    {
	int i;
	jdouble* pb;
	ary = (*jenv)->NewDoubleArray(jenv, RARRAY(v)->len);
	pb = (*jenv)->GetDoubleArrayElements(jenv, ary, NULL);
	for (i = 0; i < RARRAY(v)->len; i++)
	{
	    *(pb + i) = (jdouble)rb_num2dbl(RARRAY(v)->ptr[i]);
	}
	(*jenv)->ReleaseDoubleArrayElements(jenv, ary, pb, 0);
	return ary;
    }
    rb_raise(rb_eRuntimeError, "can't coerce to double array");
}

static jarray r2farray(VALUE v, const char* cls)
{
    jarray ary = NULL;
    if (TYPE(v) == T_ARRAY)
    {
	int i;
	jfloat* pb;
	ary = (*jenv)->NewFloatArray(jenv, RARRAY(v)->len);
	pb = (*jenv)->GetFloatArrayElements(jenv, ary, NULL);
	for (i = 0; i < RARRAY(v)->len; i++)
	{
	    *(pb + i) = (jfloat)rb_num2dbl(RARRAY(v)->ptr[i]);
	}
	(*jenv)->ReleaseFloatArrayElements(jenv, ary, pb, 0);
	return ary;
    }
    rb_raise(rb_eRuntimeError, "can't coerce to float array");
}

static jarray r2iarray(VALUE v, const char* cls)
{
    jarray ary = NULL;
    if (TYPE(v) == T_ARRAY)
    {
	int i;
	jint* pb;
	check_fixnumarray(v);
	ary = (*jenv)->NewIntArray(jenv, RARRAY(v)->len);
	pb = (*jenv)->GetIntArrayElements(jenv, ary, NULL);
	for (i = 0; i < RARRAY(v)->len; i++)
	{
	    *(pb + i) = (jint)FIX2LONG(RARRAY(v)->ptr[i]);
	}
	(*jenv)->ReleaseIntArrayElements(jenv, ary, pb, 0);
	return ary;
    }
    rb_raise(rb_eRuntimeError, "can't coerce to int array");
}

static jarray r2larray(VALUE v, const char* cls)
{
    jarray ary = NULL;
    if (TYPE(v) == T_ARRAY)
    {
	int i;
	jlong* pb;
	ary = (*jenv)->NewLongArray(jenv, RARRAY(v)->len);
	pb = (*jenv)->GetLongArrayElements(jenv, ary, NULL);
	for (i = 0; i < RARRAY(v)->len; i++)
	{
#if HAVE_LONG_LONG
	    *(pb + i) = (jlong)rb_num2ll(RARRAY(v)->ptr[i]);
#else
	    *(pb + i) = (jlong)FIX2LONG(RARRAY(v)->ptr[i]);
#endif
	}
	(*jenv)->ReleaseLongArrayElements(jenv, ary, pb, 0);
	return ary;
    }
    rb_raise(rb_eRuntimeError, "can't coerce to long array");
}

static jarray r2sarray(VALUE v, const char* cls)
{
    jarray ary = NULL;
    if (TYPE(v) == T_ARRAY)
    {
	int i;
	jshort* pb;
	check_fixnumarray(v);
	ary = (*jenv)->NewShortArray(jenv, RARRAY(v)->len);
	pb = (*jenv)->GetShortArrayElements(jenv, ary, NULL);
	for (i = 0; i < RARRAY(v)->len; i++)
	{
	    *(pb + i) = (jshort)FIX2LONG(RARRAY(v)->ptr[i]);
	}
	(*jenv)->ReleaseShortArrayElements(jenv, ary, pb, 0);
	return ary;
    }
    rb_raise(rb_eRuntimeError, "can't coerce to short array");
}

static jarray r2boolarray(VALUE v, const char* cls)
{
    jarray ary = NULL;
    if (TYPE(v) == T_ARRAY)
    {
	int i;
	jboolean* pb;
	ary = (*jenv)->NewBooleanArray(jenv, RARRAY(v)->len);
	pb = (*jenv)->GetBooleanArrayElements(jenv, ary, NULL);
	for (i = 0; i < RARRAY(v)->len; i++)
	{
	    *(pb + i)
		= (NIL_P(RARRAY(v)->ptr[i]) || RARRAY(v)->ptr[i] == Qfalse)
			? JNI_FALSE : JNI_TRUE;
	}
	(*jenv)->ReleaseBooleanArrayElements(jenv, ary, pb, 0);
	return ary;
    }
    rb_raise(rb_eRuntimeError, "can't coerce to boolean array");
}

static jarray r2voidarray(VALUE v, const char* cls)
{
    rb_raise(rb_eRuntimeError, "void never arrayed");
}

static jarray r2objarray(VALUE v, const char* cls)
{
    jarray ary = NULL;
    if (TYPE(v) == T_ARRAY)
    {
	int i;
	ary = (*jenv)->NewObjectArray(jenv, RARRAY(v)->len, j_object, NULL);
	check_exception(0);
	for (i = 0; i < RARRAY(v)->len; i++)
	{
	    VALUE val = RARRAY(v)->ptr[i];
	    if (TYPE(val) == T_DATA
		&& (RBASIC(val)->klass == rjbi || RBASIC(val)->klass == rjb))
	    {
		struct jvi_data* ptr;
		Data_Get_Struct(val, struct jvi_data, ptr);
		(*jenv)->SetObjectArrayElement(jenv, ary, i, ptr->obj);
	    }
	    else
	    {
		(*jenv)->SetObjectArrayElement(jenv, ary, i, NULL);
	    }
	}
	return ary;
    }
    rb_raise(rb_eRuntimeError, "can't coerce to object array");
}

static const J2R jarray2r[] = {
    jbytearray2rv, jchararray2rv, jdoublearray2rv, jfloatarray2rv,
    jintarray2rv, jlongarray2rv, jshortarray2rv, jbooleanarray2rv,
    NULL /* void */, jstringarray2rv,
};

static const R2JARRAY r2jarray[] = {
    r2barray, r2carray, r2darray, r2farray, r2iarray, r2larray, r2sarray,
    r2boolarray, r2voidarray, r2objarray,
};

static void rv2jarray(VALUE val, jvalue* jv, const char* psig, int release)
{
    if (*psig != '[')
    {
	rb_raise(rb_eRuntimeError, "argument signature not array");
    }
    if (*(psig + 1) == '[')
    {
	rb_raise(rb_eRuntimeError, "only unarray is supported");
    }
    if (release)
    {
	(*jenv)->DeleteLocalRef(jenv, jv->l);
    }
    else
    {
	int i;
	R2JARRAY r2a = r2objarray;
	for (i = 0; i < COUNTOF(r2jarray); i++)
	{
	    if (*(psig + 1) == jtype_etbl[i][0])
	    {
		r2a = r2jarray[i];
		break;
	    }
	}
	jv->l = r2a(val, psig + 1);
    }
}

static off_t jcall[] = {
    offsetof(struct JNINativeInterface_, CallByteMethodA),
    offsetof(struct JNINativeInterface_, CallCharMethodA),    
    offsetof(struct JNINativeInterface_, CallDoubleMethodA),
    offsetof(struct JNINativeInterface_, CallFloatMethodA),
    offsetof(struct JNINativeInterface_, CallIntMethodA),
    offsetof(struct JNINativeInterface_, CallLongMethodA),
    offsetof(struct JNINativeInterface_, CallShortMethodA),
    offsetof(struct JNINativeInterface_, CallBooleanMethodA),
    offsetof(struct JNINativeInterface_, CallVoidMethodA),
    offsetof(struct JNINativeInterface_, CallObjectMethodA),    
};

static off_t jscall[] = {
    offsetof(struct JNINativeInterface_, CallStaticByteMethodA),
    offsetof(struct JNINativeInterface_, CallStaticCharMethodA),    
    offsetof(struct JNINativeInterface_, CallStaticDoubleMethodA),
    offsetof(struct JNINativeInterface_, CallStaticFloatMethodA),
    offsetof(struct JNINativeInterface_, CallStaticIntMethodA),
    offsetof(struct JNINativeInterface_, CallStaticLongMethodA),
    offsetof(struct JNINativeInterface_, CallStaticShortMethodA),
    offsetof(struct JNINativeInterface_, CallStaticBooleanMethodA),
    offsetof(struct JNINativeInterface_, CallStaticVoidMethodA),
    offsetof(struct JNINativeInterface_, CallStaticObjectMethodA),    
};

/*
 */
static void get_j2r(jobject cls, struct cls_method* pm, int static_method)
{
    int i;
    off_t iv = 0;
    J2R result = NULL;
    const char* cname;
    jstring nm = (*jenv)->CallObjectMethod(jenv, cls, class_getName);
    check_exception(0);	    
    cname = (*jenv)->GetStringUTFChars(jenv, nm, JNI_FALSE);

    if (*cname == '[')
    {
	if (*(cname + 1) == '[')
	{
	    rb_raise(rb_eRuntimeError, "only unarray is supported");
	}
	for (i = 0; i < COUNTOF(jtype_etbl); i++)
	{
	    if (*(cname + 1) == jtype_etbl[i][0])
	    {
		if (jtype_etbl[i][0] == 'L'
		    && strncmp(cname + 1, jtype_etbl[i], strlen(jtype_etbl[i])))
		{
		    break;
		}
		result = jarray2r[i];
		break;
	    }
	}
	if (result == NULL)
	{
	    result = jarray2rv;
	}
    }
    else
    {
	for (i = 0; i < COUNTOF(jtype_tbl); i++)
	{
	    if (!strcmp(cname, jtype_tbl[i]))
	    {
		result = j2r[i];
		iv = (static_method) ? jscall[i] : jcall[i];
		break;
	    }
	}
    }
    (*jenv)->ReleaseStringUTFChars(jenv, nm, cname);
    (*jenv)->DeleteLocalRef(jenv, nm);
    pm->result_convert = (result) ? result : jv2rv;
    if (iv)
    {
	pm->method = iv;
    }
    else
    {
	pm->method = (static_method)
	    ? offsetof(struct JNINativeInterface_, CallStaticObjectMethodA)
	    : offsetof(struct JNINativeInterface_, CallObjectMethodA);
    }
}

static void fill_convert(struct cls_constructor* cls, jobjectArray tp, int count)
{
    int i, j, siglen, len;
    R2J* tbl = ALLOC_N(R2J, count);
    const char** sig = ALLOCA_N(char*, count);
    cls->arg_convert = tbl;
    memset(tbl, 0, sizeof(R2J) * count);
    siglen = 0;
    
    for (i = 0; i < count; i++)
    {
	jstring nm;
	const char* cname;
	jobject o = (*jenv)->GetObjectArrayElement(jenv, tp, i);
	nm = (*jenv)->CallObjectMethod(jenv, o, class_getName);
	check_exception(0);	    
	cname = (*jenv)->GetStringUTFChars(jenv, nm, JNI_FALSE);
	if (*cname == '[')
	{
	    len = strlen(cname);
	    siglen += len;
	    *(sig + i) = ALLOC_N(char, len + 1);
	    strcpy(*(char**)(sig + i), cname);
	    *(tbl + i) = rv2jarray;
	}
	else
	{
	    for (j = 0; j < COUNTOF(jtype_tbl); j++)
	    {
		if (!strcmp(cname, jtype_tbl[j]))
		{
		    *(tbl + i) = r2j[j];
		    *(sig + i) = jtype_etbl[j];
		    siglen += strlen(jtype_etbl[j]);
		    break;
		}
	    }
	    if (*(tbl + i) == NULL)
	    {
		*(tbl + i) = rv2jobject;
		*(sig + i) = ALLOCA_N(char, strlen(cname) + 3);
		siglen += sprintf(*(char**)(sig + i), "L%s;", cname);
	    }
	}
	(*jenv)->ReleaseStringUTFChars(jenv, nm, cname);
	(*jenv)->DeleteLocalRef(jenv, nm);
	if (!*(tbl + i))
	{
	    *(tbl + i) = rv2jv;
	}
    }
    if (siglen > 0)
    {
	cls->method_signature = ALLOC_N(char, siglen + 1);
	*(cls->method_signature) = 0;
	for (i = 0; i < count; i++)
	{
	    strcat(cls->method_signature, *(sig + i));
	}
    }
}

/*
 * create method info structure
 * m = instance of Method class
 * c = instance of the class
 */
static void setup_methodbase(struct cls_constructor* pm,
			     jobjectArray parama, jsize pcount)
{
    pm->arg_count = pcount;
    pm->method_signature = NULL;
    pm->arg_convert = NULL;
    if (pcount)
    {
	fill_convert(pm, parama, pcount);
    }
}
 
static void create_methodinfo(st_table* tbl, jobject m, int static_method)
{
    struct cls_method* result;
    struct cls_method* pm;    
    char* param = NULL;
    const char* jname;
    jstring nm;
    jobjectArray parama;
    jobject cls;
    jsize param_count;

    result = ALLOC(struct cls_method);
    parama = (*jenv)->CallObjectMethod(jenv, m, getParameterTypes);
    check_exception(0);
    param_count = (*jenv)->GetArrayLength(jenv, parama);
    check_exception(0);
    setup_methodbase(&result->basic, parama, param_count);

    nm = (*jenv)->CallObjectMethod(jenv, m, method_getName);
    check_exception(0);	    
    jname = (*jenv)->GetStringUTFChars(jenv, nm, JNI_FALSE);
    result->name = rb_intern(jname);
    (*jenv)->ReleaseStringUTFChars(jenv, nm, jname);
    (*jenv)->DeleteLocalRef(jenv, nm);
    result->basic.id = (*jenv)->FromReflectedMethod(jenv, m);
    check_exception(0);
    cls = (*jenv)->CallObjectMethod(jenv, m, getReturnType);
    check_exception(0);
    get_j2r(cls, result, static_method);
    result->static_method = static_method;
    if (st_lookup(tbl, result->name, (st_data_t*)&pm))
    {
	result->next = pm->next;
	pm->next = result;
    }
    else
    {
	result->next = NULL;
	st_insert(tbl, result->name, (VALUE)result);
    }
}

static void setup_constructors(struct cls_constructor*** pptr, jobjectArray methods)
{
    int i;
    struct cls_constructor* pc;    
    jsize mcount = (*jenv)->GetArrayLength(jenv, methods);
    struct cls_constructor** tbl = ALLOC_N(struct cls_constructor*, mcount + 1);
    *pptr = tbl;
    for (i = 0; i < mcount; i++)
    {
	jobjectArray parama;
	jsize pcount;
	jobject c = (*jenv)->GetObjectArrayElement(jenv, methods, i);
	check_exception(0);
	pc = ALLOC(struct cls_constructor);
	tbl[i] = pc;
	parama = (*jenv)->CallObjectMethod(jenv, c, ctrGetParameterTypes);
	check_exception(0);
	pcount = (*jenv)->GetArrayLength(jenv, parama);
	check_exception(0);
	setup_methodbase(pc, parama, pcount);
	pc->id = (*jenv)->FromReflectedMethod(jenv, c);
    }
    tbl[mcount] = NULL;    
}
			       
static void setup_methods(st_table** tbl, jobjectArray methods)
{
    int i;
    jint modifier;    
    jsize mcount = (*jenv)->GetArrayLength(jenv, methods);
    *tbl = st_init_numtable_with_size(mcount);
    for (i = 0; i < mcount; i++)
    {
	jobject m = (*jenv)->GetObjectArrayElement(jenv, methods, i);
	check_exception(0);	
	modifier = (*jenv)->CallIntMethod(jenv, m, method_getModifiers);
	if (!(modifier & ACC_STATIC))
	{
	    create_methodinfo(*tbl, m, 0);
	}
    }
}

static void setup_staticmethods(st_table** tbl, jobjectArray methods)
{
    int i;
    jint modifier;
    jsize mcount = (*jenv)->GetArrayLength(jenv, methods);
    *tbl = st_init_numtable();
    for (i = 0; i < mcount; i++)
    {
	jobject m = (*jenv)->GetObjectArrayElement(jenv, methods, i);
	check_exception(0);
	modifier = (*jenv)->CallIntMethod(jenv, m, method_getModifiers);
	check_exception(0);
	if ((modifier & (ACC_PUBLIC | ACC_STATIC)) == (ACC_PUBLIC | ACC_STATIC))
	{
	    create_methodinfo(*tbl, m, 1);
	}
    }
}

static void setup_metadata(struct jv_data* ptr)
{
    jmethodID mid;
    jobjectArray methods;
    
    jclass klass;
    
    klass = (*jenv)->GetObjectClass(jenv, ptr->idata.obj);
    ptr->idata.klass = klass;
    check_exception(0);
    mid = (*jenv)->GetMethodID(jenv, klass, "getMethods", "()[Ljava/lang/reflect/Method;");
    check_exception(0);
    methods = (*jenv)->CallNonvirtualObjectMethod(jenv, ptr->idata.obj, klass, mid);
    check_exception(0);
    setup_methods(&ptr->idata.methods, methods);
    mid = (*jenv)->GetMethodID(jenv, klass, "getDeclaredMethods", "()[Ljava/lang/reflect/Method;");
    check_exception(0);
    methods = (*jenv)->CallNonvirtualObjectMethod(jenv, ptr->idata.obj, klass, mid);
    check_exception(0);
    setup_staticmethods(&ptr->static_methods, methods);
    mid = (*jenv)->GetMethodID(jenv, klass, "getConstructors", "()[Ljava/lang/reflect/Constructor;");
    check_exception(0);
    methods = (*jenv)->CallNonvirtualObjectMethod(jenv, ptr->idata.obj, klass, mid);
    check_exception(0);
    setup_constructors(&ptr->constructors, methods);
}

/*
 * load Java Virtual Machine
 * def load(version = 1.4, class_path = '')
 */
static VALUE rjb_s_load(int argc, VALUE* argv, VALUE self)
{
    jint res;
    JDK1_1InitArgs vm_args;
    char* userpath;
    char* newpath;
    int len;

    if (jvm != NULL)
    {
	return Qnil;
    }

    memset(&vm_args, 0, sizeof(vm_args));
    if (argc < 1)
    {
	vm_args.version = JNI_VERSION_1_4;
    }
    else
    {
	rb_check_type(*argv, T_FLOAT);
	vm_args.version = (1 << 16) | (int)((RFLOAT(*argv)->value) * 10);
    }
    JNI_GetDefaultJavaVMInitArgs(&vm_args);
    if (argc >= 2)
    {
	rb_check_type(*(argv + 1), T_STRING);
	userpath = StringValueCStr(argv[1]);
    }
    else
    {
	userpath = ".";
    }
    len = strlen(userpath);
    if (vm_args.classpath)
    {
	newpath = ALLOCA_N(char, len + strlen(vm_args.classpath) + 2);
	sprintf(newpath, "%s%c%s", userpath, PATH_SEP, vm_args.classpath);
    }
    else
    {
	newpath = userpath;
    }
    vm_args.classpath = newpath;

    res = JNI_CreateJavaVM(&jvm, (void**)&jenv, &vm_args);
    if (res < 0)
    {
	jvm = NULL;
	rb_raise(rb_eRuntimeError, "can't create Java VM");
    }

    j_constructor = (*jenv)->FindClass(jenv, "java/lang/reflect/Constructor");
    check_exception(1);
    ctrGetParameterTypes = (*jenv)->GetMethodID(jenv, j_constructor, "getParameterTypes", "()[Ljava/lang/Class;");
    check_exception(1);
    j_method = (*jenv)->FindClass(jenv, "java/lang/reflect/Method");
    method_getModifiers = (*jenv)->GetMethodID(jenv, j_method, "getModifiers", "()I");
    check_exception(1);
    method_getName = (*jenv)->GetMethodID(jenv, j_method, "getName", "()Ljava/lang/String;");
    check_exception(1);	
    getParameterTypes = (*jenv)->GetMethodID(jenv, j_method, "getParameterTypes", "()[Ljava/lang/Class;");
    check_exception(1);	
    getReturnType = (*jenv)->GetMethodID(jenv, j_method, "getReturnType", "()Ljava/lang/Class;");
    check_exception(1);

    j_class = (*jenv)->FindClass(jenv, "java/lang/Class");
    check_exception(1);
    class_getName = (*jenv)->GetMethodID(jenv, j_class, "getName", "()Ljava/lang/String;");
    check_exception(1);

    j_string = (*jenv)->FindClass(jenv, "java/lang/String");
    check_exception(1);
    str_tostring = (*jenv)->GetMethodID(jenv, j_string, "toString", "()Ljava/lang/String;");
    check_exception(1);

    j_object = (*jenv)->FindClass(jenv, "java/lang/Object");
    check_exception(1);    
    
    return Qnil;
}

/*
 * unload Java Virtual Machine
 *
 * def unload()
 *   classes.clear
 *   unload(jvm)
 * end
 */
static int clear_classes(VALUE key, VALUE val, VALUE dummy)
{
    return ST_DELETE;
}
static VALUE rjb_s_unload(int argc, VALUE* argv, VALUE self)
{
    st_foreach(RHASH(loaded_classes)->tbl, clear_classes, 0);
    if (jvm)
    {
	(*jvm)->DestroyJavaVM(jvm);
	jvm = NULL;
    }
    return Qnil;
}

/*
 * return all classes that were already loaded.
 * this method simply returns the global hash,
 * but it's safe because the hash was frozen.
 */
static VALUE rjb_s_classes(VALUE self)
{
    return loaded_classes;
}

/*
 * free java class
 */
static void free_constructor(struct cls_constructor* p)
{
    free(p->arg_convert);
    free(p->method_signature);
}
static int free_method_item(ID key, struct cls_method* pm, int dummy)
{
    for (; pm; pm = pm->next)
    {
	free_constructor(&pm->basic);
    }
    return ST_CONTINUE;
}

/*
 * finalize Object instance
 */
static VALUE rjb_delete_local(struct jvi_data* ptr)
{
    if (jenv)
    {
	(*jenv)->DeleteLocalRef(jenv, ptr->obj);
    }
    return Qnil;
}

/*
 * finalize Class instance
 */
static VALUE rjb_s_free(struct jv_data* ptr)
{
    /* class never delete
    struct cls_constructor** c;
    
    rjb_delete_local(&ptr->idata);
    if (ptr->constructors)
    {
	for (c = ptr->constructors; *c; c++)
	{
	    free_constructor(*c);
	}
    }
    free(ptr->constructors);
    if (ptr->idata.methods)
    {
	st_foreach(ptr->idata.methods, (int(*)())free_method_item, 0);
	st_free_table(ptr->idata.methods);
    }
    (*jenv)->DeleteLocalRef(jenv, ptr->klass);	
    st_delete(RHASH(loaded_classes)->tbl, clsname, NULL);
    */
    return Qnil;
}

/*
 * create new instance of this class
 */
static VALUE createinstance(int argc, VALUE* argv,
	    struct jvi_data* org, struct cls_constructor* pc)
{
    int i;
    char* psig = pc->method_signature;
    jobject obj = NULL;
    jvalue* args = (argc) ? ALLOCA_N(jvalue, argc) : NULL;
    for (i = 0; i < argc; i++)
    {
	R2J pr2j = *(pc->arg_convert + i);
	pr2j(argv[i], args + i, psig, 0);
	psig = next_sig(psig);
    }
    obj = (*jenv)->NewObjectA(jenv, org->obj, pc->id, args);
    if (!obj)
    {
	check_exception(1);
    }
    psig = pc->method_signature;
    for (i = 0; i < argc; i++)
    {
	R2J pr2j = *(pc->arg_convert + i);
	pr2j(argv[i], args + i, psig, 1);
	psig = next_sig(psig);
    }

    return register_instance(org, obj);
}

static VALUE register_instance(struct jvi_data* org, jobject obj)
{
    struct jvi_data* ptr = ALLOC(struct jvi_data);
    VALUE v = Data_Wrap_Struct(rjbi, NULL, rjb_delete_local, ptr);
    ptr->klass = org->obj;
    ptr->obj = obj;
    ptr->methods = org->methods;
    return v;
}

/*
 * new instance with signature
 */
static VALUE rjb_newinstance_s(int argc, VALUE* argv, VALUE self)
{
    VALUE vsig, rest;
    char* sig;
    VALUE ret = Qnil;
    struct jv_data* ptr;

    rb_scan_args(argc, argv, "1*", &vsig, &rest);
    sig = StringValueCStr(vsig);
    Data_Get_Struct(self, struct jv_data, ptr);
    if (ptr->constructors)
    {
	struct cls_constructor** pc = ptr->constructors;
	for (pc = ptr->constructors; *pc; pc++)
	{
	    if ((*pc)->arg_count == argc - 1
		&& !strcmp(sig, (*pc)->method_signature))
	    {
		ret = createinstance(argc - 1, argv + 1, &ptr->idata, *pc);
		break;
	    }
	}
    }
    return ret;
}

static VALUE rjb_newinstance(int argc, VALUE* argv, VALUE self)
{
    VALUE ret = Qnil;
    struct jv_data* ptr;
    struct cls_constructor** pc;
    Data_Get_Struct(self, struct jv_data, ptr);

    if (ptr->constructors)
    {
	for (pc = ptr->constructors; *pc; pc++)
	{
	    if ((*pc)->arg_count == argc)
	    {
		ret = createinstance(argc, argv, &ptr->idata, *pc);
		break;
	    }
	}
    }
    return ret;
}

/*
 * Jclass Rjb::import(classname)
 */
static VALUE rjb_s_import(VALUE self, VALUE clsname)
{
    struct jv_data* ptr;
    char* cname;
    char* jnicls;
    char* p;
    VALUE v = rb_hash_aref(loaded_classes, clsname);
    if (v != Qnil)
    {
	return v;
    }

    ptr = ALLOC(struct jv_data);
    memset(ptr, 0, sizeof(struct jv_data));
    v = Data_Wrap_Struct(rjbc, NULL, rjb_s_free, ptr);
    
    cname = StringValueCStr(clsname);
    jnicls = ALLOCA_N(char, strlen(cname) + 1);
    strcpy(jnicls, cname);
    for (p = jnicls; *p; p++)
    {
	if (*p == '.')
	{
	    *p = '/';
	}
    }
    ptr->idata.obj = (*jenv)->FindClass(jenv, jnicls);
    if (!ptr->idata.obj)
    {
	check_exception(0);
	rb_raise(rb_eRuntimeError, "`%s' not found", cname);
    }
    setup_metadata(ptr);
    register_class(v, clsname);
    return v;
}

static void register_class(VALUE self, VALUE clsname)
{
    rb_define_singleton_method(self, "new", rjb_newinstance, -1);
    rb_define_singleton_method(self, "new_with_sig", rjb_newinstance_s, -1);
    /*
     * the hash was frozen, so it need to call st_ func directly.
     */
    st_insert(RHASH(loaded_classes)->tbl, clsname, self);
}

/*
 * temporary signature
 */
static const char* check_rtype(VALUE v)
{
    switch (TYPE(v))
    {
    case T_FIXNUM:
	return "I";
    case T_FLOAT:
	return "D";
    case T_STRING:
	return "Ljava.lang.String;";
    case T_TRUE:
    case T_FALSE:
	return "Z";
    case T_OBJECT:
	// fall down to the next case
    default:
	return NULL;
    }
}

/*
 * return class name
 */
static VALUE rjb_i_class(VALUE self)
{
    struct jvi_data* ptr;
    jstring nm;
    Data_Get_Struct(self, struct jvi_data, ptr);
    nm = (*jenv)->CallObjectMethod(jenv, ptr->klass, class_getName);
    check_exception(0);
    return jstring2val(nm);
}

/*
 * invoker
 */
static VALUE invoke(struct cls_method* pm, struct jvi_data* ptr,
		    int argc, VALUE* argv, const char* sig)
{
    int i, found;
    jvalue jv;
    jvalue* args;
    INVOKEA  voke;
    char* psig;
    /* for future 
    if (sig == NULL && argc)
    {
	int len;
	char** sigs = ALLOCA_N(char*, argc);
	memset(*sigs, 0, sizeof(char*) * argc);
	for (i = 0; i < argc; i++)
	{
	    const char* typ = check_rtype(*(argv + i));
	    if (!typ) break;
	    *(sigs + i) = ALLOCA_N(char, strlen(typ) + 1);
	    strcpy(*(sigs + i), typ);
	}
	for (i = 0; i < argc; i++)
	{
	    if (!*(sigs + i))
	    {
		len = -1;
		break;
	    }
	    len += strlen(*(sigs + i));
	}
	if (len > 0)
	{
	    s = ALLOCA_N(char, len + 1);
	    *s = 0;
	    for (i = 0; i < argc; i++)
	    {
		strcat(s, *(sigs + i));
	    }
	}
    }
    */
    for (found = 0; pm; pm = pm->next)
    {
	if (argc == pm->basic.arg_count
	    && (!sig || !strcmp(sig, pm->basic.method_signature)))
	{
	    found = 1;
	    break;
	}
    }
    if (!found)
    {
	char* tname = rb_id2name(pm->name);
	rb_raise(rb_eRuntimeError, "Fail: unknown method name `%s'", tname);
    }
    voke = *(INVOKEA*)(((char*)*jenv) + pm->method);
    args = (argc) ? ALLOCA_N(jvalue, argc) : NULL;
    psig = pm->basic.method_signature;
    for (i = 0; i < argc; i++)
    {
	R2J pr2j = *(pm->basic.arg_convert + i);
	pr2j(argv[i], args + i, psig, 0);
	psig = next_sig(psig);
    }
    jv = voke(jenv, ptr->obj, pm->basic.id, args);
    check_exception(1);
    psig = pm->basic.method_signature;
    for (i = 0; i < argc; i++)
    {
	R2J pr2j = *(pm->basic.arg_convert + i);
	pr2j(argv[i], args + i, psig, 1);
	psig = next_sig(psig);
    }
    return pm->result_convert(jv);
}

/*
 * Object invocation
 */
static VALUE invoke_by_instance(ID rmid, int argc, VALUE* argv,
				struct jvi_data* ptr, char* sig)
{
    VALUE ret = Qnil;
    struct cls_method* pm;
    
    if (st_lookup(ptr->methods, rmid, (st_data_t*)&pm))
    {
	ret = invoke(pm, ptr, argc, argv, sig);
    }
    else
    {
	char* tname = rb_id2name(rmid);
	rb_raise(rb_eRuntimeError, "Fail: unknown method name `%s'", tname);
    }
    return ret;
}

static VALUE rjb_i_invoke(int argc, VALUE* argv, VALUE self)
{
    VALUE vsig, rmid, rest;
    char* sig;
    struct jvi_data* ptr;

    rb_scan_args(argc, argv, "2*", &rmid, &vsig, &rest);
    rmid = rb_to_id(rmid);
    sig = StringValueCStr(vsig);
    Data_Get_Struct(self, struct jvi_data, ptr);

    return invoke_by_instance(rmid, argc -2, argv + 2, ptr, sig);
}

static VALUE rjb_i_missing(int argc, VALUE* argv, VALUE self)
{
    struct jvi_data* ptr;
    ID rmid = rb_to_id(argv[0]);
    
    Data_Get_Struct(self, struct jvi_data, ptr);
    
    return invoke_by_instance(rmid, argc -1, argv + 1, ptr, NULL);
}

/*
 * Class invocation (first static method, then instance method)
 */
static VALUE invoke_by_class(ID rmid, int argc, VALUE* argv,
			     struct jv_data* ptr, char* sig)
{
    VALUE ret = Qnil;
    struct cls_method* pm;

    if (st_lookup(ptr->static_methods, rmid, (st_data_t*)&pm))
    {
	ret = invoke(pm, &ptr->idata, argc, argv, sig);
    }
    else if (st_lookup(ptr->idata.methods, rmid, (st_data_t*)&pm))
    {
	ret = invoke(pm, &ptr->idata, argc, argv, sig);
    }
    else
    {
	char* tname = rb_id2name(rmid);
	rb_raise(rb_eRuntimeError, "Fail: unknown method name `%s'", tname);
    }
    
    return ret;
}

static VALUE rjb_invoke(int argc, VALUE* argv, VALUE self)
{
    VALUE vsig, rmid, rest;
    char* sig;
    struct jv_data* ptr;
    
    rb_scan_args(argc, argv, "2*", &rmid, &vsig, &rest);
    rmid = rb_to_id(rmid);
    sig = StringValueCStr(vsig);
    Data_Get_Struct(self, struct jv_data, ptr);

    return invoke_by_class(rmid, argc - 2, argv + 2, ptr, sig);
}

static VALUE rjb_missing(int argc, VALUE* argv, VALUE self)
{
    struct jv_data* ptr;
    ID rmid = rb_to_id(argv[0]);
    
    Data_Get_Struct(self, struct jv_data, ptr);
    return invoke_by_class(rmid, argc - 1, argv + 1, ptr, NULL);
}

/*
 * Class initializer called by Ruby while requiring this library
 */
void Init_rjb()
{
    loaded_classes = rb_hash_new();
    OBJ_FREEZE(loaded_classes);
    rb_global_variable(&loaded_classes);

    rjb = rb_define_module("Rjb");
    rb_define_module_function(rjb, "load", rjb_s_load, -1);
    rb_define_module_function(rjb, "unload", rjb_s_unload, -1);
    rb_define_module_function(rjb, "import", rjb_s_import, 1);
    rb_define_const(rjb, "VERSION", rb_str_new2(RJB_VERSION));

    /* Java class object */    
    rjbc = rb_class_new(rb_cObject);
    rb_define_method(rjbc, "method_missing", rjb_missing, -1);
    rb_define_method(rjbc, "_invoke", rjb_invoke, -1);
    rb_define_method(rjbc, "_classname", rjb_i_class, 0);    

    rb_define_singleton_method(rjbc, "classes", rjb_s_classes, 0);

    /* Java instance object */
    rjbi = rb_class_new(rb_cObject);
    rb_define_method(rjbi, "method_missing", rjb_i_missing, -1);    
    rb_define_method(rjbi, "_invoke", rjb_i_invoke, -1);
    rb_define_method(rjbi, "_classname", rjb_i_class, 0);
}
